/*
 * Copyright (c) 2018 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */

/** @file
 *  @brief Nordic UART Service Client sample
 */

#include <errno.h>
#include <zephyr/kernel.h>
#include <zephyr/device.h>
#include <zephyr/devicetree.h>
#include <zephyr/sys/byteorder.h>
#include <zephyr/sys/printk.h>

#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/hci.h>
#include <zephyr/bluetooth/conn.h>
#include <zephyr/bluetooth/uuid.h>
#include <zephyr/bluetooth/gatt.h>

#include <bluetooth/services/nus.h>
#include <bluetooth/services/nus_client.h>
#include <bluetooth/gatt_dm.h>
#include <bluetooth/scan.h>

#include <zephyr/settings/settings.h>

#include <zephyr/drivers/uart.h>

#include <zephyr/logging/log.h>
#include <zephyr/drivers/gpio.h>
#include <hal/nrf_gpio.h>

#include <zephyr/bluetooth/hci.h>
#include <zephyr/bluetooth/hci_vs.h>
#include <zephyr/sys/byteorder.h>

#define MONITOR_TXPOWER_SETTING 21

#define LOG_MODULE_NAME central_uart
LOG_MODULE_REGISTER(LOG_MODULE_NAME);

/* UART payload buffer element size. */
#define UART_BUF_SIZE 20

#define KEY_PASSKEY_ACCEPT DK_BTN1_MSK
#define KEY_PASSKEY_REJECT DK_BTN2_MSK

#define NUS_WRITE_TIMEOUT K_MSEC(150)
#define UART_WAIT_FOR_BUF_DELAY K_MSEC(50)
#define UART_RX_TIMEOUT 50

/////
//#define BT840 840
//#define BT840N_V3 8403	//52840+21540
#define BT840N_V4 8404	//52840+21540
//#define nRF21540DK 2540	

#define MONITOR_TXPOWER_SETTING 21

#define client_sent_test
#ifdef client_sent_test
	struct k_timer my_timer;
	#define TIMER_INTERVAL_MSEC 500
	uint8_t count_val = 0;
#endif

static bool is_connect = false;

const struct device *led_dev;
const struct device *led_dev1;
/////

static const struct device *uart = DEVICE_DT_GET(DT_NODELABEL(uart0));
static struct k_work_delayable uart_work;

K_SEM_DEFINE(nus_write_sem, 0, 1);

struct uart_data_t {
	void *fifo_reserved;
	uint8_t  data[UART_BUF_SIZE];
	uint16_t len;
};

#ifdef client_sent_test
	#if 0
	struct client_sent_buf_t {
  		uint8_t m_tx_buf[100];
	};
	#endif
#endif

////////////////
uint8_t	Dummy_tx_buf[15];	
uint16_t Dummy_len;	
uint8_t Dummy_cnt= 0x30;
////////////////

static K_FIFO_DEFINE(fifo_uart_tx_data);
static K_FIFO_DEFINE(fifo_uart_rx_data);

static struct bt_conn *default_conn;
static struct bt_nus_client nus_client;

static void ble_data_sent(struct bt_nus_client *nus, uint8_t err,
					const uint8_t *const data, uint16_t len)
{
	ARG_UNUSED(nus);

	//struct uart_data_t *buf;

	/* Retrieve buffer context. */
/*
	buf = CONTAINER_OF(data, struct uart_data_t, data);
	k_free(buf);

	k_sem_give(&nus_write_sem);

	if (err) {
		LOG_WRN("ATT error code: 0x%02X", err);
	}
*/
}

static uint8_t ble_data_received(struct bt_nus_client *nus,
						const uint8_t *data, uint16_t len)
{
	ARG_UNUSED(nus);

	int err;

	for (uint16_t pos = 0; pos != len;) {
		struct uart_data_t *tx = k_malloc(sizeof(*tx));

		if (!tx) {
			LOG_WRN("Not able to allocate UART send data buffer");
			return BT_GATT_ITER_CONTINUE;
		}

		/* Keep the last byte of TX buffer for potential LF char. */
		size_t tx_data_size = sizeof(tx->data) - 1;

		if ((len - pos) > tx_data_size) {
			tx->len = tx_data_size;
		} else {
			tx->len = (len - pos);
		}

		memcpy(tx->data, &data[pos], tx->len);

		pos += tx->len;

		/* Append the LF character when the CR character triggered
		 * transmission from the peer.
		 */
		if ((pos == len) && (data[len - 1] == '\r')) {
			tx->data[tx->len] = '\n';
			tx->len++;
		}

		err = uart_tx(uart, tx->data, tx->len, SYS_FOREVER_MS);
		if (err) {
			k_fifo_put(&fifo_uart_tx_data, tx);
		}
	}

	return BT_GATT_ITER_CONTINUE;
}

static void uart_cb(const struct device *dev, struct uart_event *evt, void *user_data)
{
	ARG_UNUSED(dev);

	static size_t aborted_len;
	struct uart_data_t *buf;
	static uint8_t *aborted_buf;
	static bool disable_req;

	switch (evt->type) {
	case UART_TX_DONE:
		LOG_DBG("UART_TX_DONE");
		if ((evt->data.tx.len == 0) ||
		    (!evt->data.tx.buf)) {
			return;
		}

		if (aborted_buf) {
			buf = CONTAINER_OF(aborted_buf, struct uart_data_t,
					   data);
			aborted_buf = NULL;
			aborted_len = 0;
		} else {
			buf = CONTAINER_OF(evt->data.tx.buf,
					   struct uart_data_t,
					   data);
		}

		k_free(buf);

		buf = k_fifo_get(&fifo_uart_tx_data, K_NO_WAIT);
		if (!buf) {
			return;
		}

		if (uart_tx(uart, buf->data, buf->len, SYS_FOREVER_MS)) {
			LOG_WRN("Failed to send data over UART");
		}

		break;

	case UART_RX_RDY:
		LOG_DBG("UART_RX_RDY");
		buf = CONTAINER_OF(evt->data.rx.buf, struct uart_data_t, data);
		buf->len += evt->data.rx.len;

		if (disable_req) {
			return;
		}

		if ((evt->data.rx.buf[buf->len - 1] == '\n') ||
		    (evt->data.rx.buf[buf->len - 1] == '\r')) {
			disable_req = true;
			uart_rx_disable(uart);
		}

		break;

	case UART_RX_DISABLED:
		LOG_DBG("UART_RX_DISABLED");
		disable_req = false;

		buf = k_malloc(sizeof(*buf));
		if (buf) {
			buf->len = 0;
		} else {
			LOG_WRN("Not able to allocate UART receive buffer");
			k_work_reschedule(&uart_work, UART_WAIT_FOR_BUF_DELAY);
			return;
		}

		uart_rx_enable(uart, buf->data, sizeof(buf->data),
			       UART_RX_TIMEOUT);

		break;

	case UART_RX_BUF_REQUEST:
		LOG_DBG("UART_RX_BUF_REQUEST");
		buf = k_malloc(sizeof(*buf));
		if (buf) {
			buf->len = 0;
			uart_rx_buf_rsp(uart, buf->data, sizeof(buf->data));
		} else {
			LOG_WRN("Not able to allocate UART receive buffer");
		}

		break;

	case UART_RX_BUF_RELEASED:
		LOG_DBG("UART_RX_BUF_RELEASED");
		buf = CONTAINER_OF(evt->data.rx_buf.buf, struct uart_data_t,
				   data);

		if (buf->len > 0) {
			k_fifo_put(&fifo_uart_rx_data, buf);
		} else {
			k_free(buf);
		}

		break;

	case UART_TX_ABORTED:
		LOG_DBG("UART_TX_ABORTED");
		if (!aborted_buf) {
			aborted_buf = (uint8_t *)evt->data.tx.buf;
		}

		aborted_len += evt->data.tx.len;
		buf = CONTAINER_OF(aborted_buf, struct uart_data_t,
				   data);

		uart_tx(uart, &buf->data[aborted_len],
			buf->len - aborted_len, SYS_FOREVER_MS);

		break;

	default:
		break;
	}
}

static void uart_work_handler(struct k_work *item)
{
	struct uart_data_t *buf;

	buf = k_malloc(sizeof(*buf));
	if (buf) {
		buf->len = 0;
	} else {
		LOG_WRN("Not able to allocate UART receive buffer");
		k_work_reschedule(&uart_work, UART_WAIT_FOR_BUF_DELAY);
		return;
	}

	uart_rx_enable(uart, buf->data, sizeof(buf->data), UART_RX_TIMEOUT);
}

static int uart_init(void)
{
	int err;
	struct uart_data_t *rx;

	if (!device_is_ready(uart)) {
		LOG_ERR("UART device not ready");
		return -ENODEV;
	}

	rx = k_malloc(sizeof(*rx));
	if (rx) {
		rx->len = 0;
	} else {
		return -ENOMEM;
	}

	k_work_init_delayable(&uart_work, uart_work_handler);

	err = uart_callback_set(uart, uart_cb, NULL);
	if (err) {
		return err;
	}

	return uart_rx_enable(uart, rx->data, sizeof(rx->data),
			      UART_RX_TIMEOUT);
}

static void discovery_complete(struct bt_gatt_dm *dm,
			       void *context)
{
	struct bt_nus_client *nus = context;
	LOG_INF("Service discovery completed");

	bt_gatt_dm_data_print(dm);

	bt_nus_handles_assign(dm, nus);
	bt_nus_subscribe_receive(nus);

	bt_gatt_dm_data_release(dm);
}

static void discovery_service_not_found(struct bt_conn *conn,
					void *context)
{
	LOG_INF("Service not found");
}

static void discovery_error(struct bt_conn *conn,
			    int err,
			    void *context)
{
	LOG_WRN("Error while discovering GATT database: (%d)", err);
}

struct bt_gatt_dm_cb discovery_cb = {
	.completed         = discovery_complete,
	.service_not_found = discovery_service_not_found,
	.error_found       = discovery_error,
};

static void gatt_discover(struct bt_conn *conn)
{
	int err;

	if (conn != default_conn) {
		return;
	}

	err = bt_gatt_dm_start(conn,
			       BT_UUID_NUS_SERVICE,
			       &discovery_cb,
			       &nus_client);
	if (err) {
		LOG_ERR("could not start the discovery procedure, error "
			"code: %d", err);
	}
}

static void exchange_func(struct bt_conn *conn, uint8_t err, struct bt_gatt_exchange_params *params)
{
	if (!err) {
		LOG_INF("MTU exchange done");
	} else {
		LOG_WRN("MTU exchange failed (err %" PRIu8 ")", err);
	}
}

static void connected(struct bt_conn *conn, uint8_t conn_err)
{
	char addr[BT_ADDR_LE_STR_LEN];
	int err;

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	if (conn_err) {
		LOG_INF("Failed to connect to %s (%d)", addr, conn_err);

		if (default_conn == conn) {
			bt_conn_unref(default_conn);
			default_conn = NULL;

			err = bt_scan_start(BT_SCAN_TYPE_SCAN_ACTIVE);
			if (err) {
				LOG_ERR("Scanning failed to start (err %d)",
					err);
			}
		}

		return;
	}

	is_connect = true;
	LOG_INF("Connected MAC Address: %s\n", addr);
	printk("Connected MAC Address: %s\n", addr);

	static struct bt_gatt_exchange_params exchange_params;

	exchange_params.func = exchange_func;
	err = bt_gatt_exchange_mtu(conn, &exchange_params);
	if (err) {
		LOG_WRN("MTU exchange failed (err %d)", err);
	}

	err = bt_conn_set_security(conn, BT_SECURITY_L2);
	if (err) {
		LOG_WRN("Failed to set security: %d", err);

		gatt_discover(conn);
	}

#ifdef nRF21540DK
	nrf_gpio_cfg_output(16);
	nrf_gpio_pin_clear(16);		//Led on
	k_sleep(K_MSEC(50));
	nrf_gpio_pin_set(16);		//Led off
	k_sleep(K_MSEC(50));
	nrf_gpio_pin_clear(16);		//Led on
	k_sleep(K_MSEC(50));
	nrf_gpio_pin_set(16);		//Led off
	k_sleep(K_MSEC(50));
	nrf_gpio_pin_clear(16);		//Led on
	k_sleep(K_MSEC(50));	
#elif BT840N_V4
	nrf_gpio_cfg_output(13);
	nrf_gpio_pin_clear(13);		//Led on
	k_sleep(K_MSEC(50));
	nrf_gpio_pin_set(13);		//Led off
	k_sleep(K_MSEC(50));
	nrf_gpio_pin_clear(13);		//Led on
	k_sleep(K_MSEC(50));
	nrf_gpio_pin_set(13);		//Led off
	k_sleep(K_MSEC(50));
	nrf_gpio_pin_clear(13);		//Led on
	k_sleep(K_MSEC(50));	
#endif

	err = bt_scan_stop();
	if ((!err) && (err != -EALREADY)) {
		LOG_ERR("Stop LE scan failed (err %d)", err);
	}
}

static void disconnected(struct bt_conn *conn, uint8_t reason)
{
	char addr[BT_ADDR_LE_STR_LEN];
	int err;

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	LOG_INF("Disconnected MAC Address: %s (reason %u)\n", addr, reason);
	printk("Disconnected MAC Address: %s (reason %u)\n", addr, reason);

	if (default_conn != conn) {
		return;
	}

	bt_conn_unref(default_conn);
	default_conn = NULL;
	is_connect = false;


#ifdef  nRF21540DK
	nrf_gpio_cfg_output(16);
	nrf_gpio_pin_set(16);		//Led off
#elif BT840N_V4
	nrf_gpio_cfg_output(13);
	nrf_gpio_pin_set(13);		//Led off
#else
#endif


	err = bt_scan_start(BT_SCAN_TYPE_SCAN_ACTIVE);
	if (err) {
		LOG_ERR("Scanning failed to start (err %d)",
			err);
	}
}

static void security_changed(struct bt_conn *conn, bt_security_t level,
			     enum bt_security_err err)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	if (!err) {
		LOG_INF("Security changed: %s level %u", addr, level);
	} else {
		LOG_WRN("Security failed: %s level %u err %d", addr,
			level, err);
	}

	gatt_discover(conn);
}

BT_CONN_CB_DEFINE(conn_callbacks) = {
	.connected = connected,
	.disconnected = disconnected,
	.security_changed = security_changed
};

static void scan_filter_match(struct bt_scan_device_info *device_info,
			      struct bt_scan_filter_match *filter_match,
			      bool connectable)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(device_info->recv_info->addr, addr, sizeof(addr));

	LOG_INF("Filters matched. Address: %s connectable: %d",
		addr, connectable);
}

static void scan_connecting_error(struct bt_scan_device_info *device_info)
{
	LOG_WRN("Connecting failed");
}

static void scan_connecting(struct bt_scan_device_info *device_info,
			    struct bt_conn *conn)
{
	default_conn = bt_conn_ref(conn);
}

static int nus_client_init(void)
{
	int err;
	struct bt_nus_client_init_param init = {
		.cb = {
			.received = ble_data_received,
			.sent = ble_data_sent,
		}
	};

	err = bt_nus_client_init(&nus_client, &init);
	if (err) {
		LOG_ERR("NUS Client initialization failed (err %d)", err);
		return err;
	}

	LOG_INF("NUS Client module initialized");
	return err;
}

BT_SCAN_CB_INIT(scan_cb, scan_filter_match, NULL,
		scan_connecting_error, scan_connecting);

static int scan_init(void)
{
	int err;
	struct bt_scan_init_param scan_init = {
		.connect_if_match = 1,
	};

	bt_scan_init(&scan_init);
	bt_scan_cb_register(&scan_cb);

	err = bt_scan_filter_add(BT_SCAN_FILTER_TYPE_UUID, BT_UUID_NUS_SERVICE);
	if (err) {
		LOG_ERR("Scanning filters cannot be set (err %d)", err);
		return err;
	}

	err = bt_scan_filter_enable(BT_SCAN_UUID_FILTER, false);
	if (err) {
		LOG_ERR("Filters cannot be turned on (err %d)", err);
		return err;
	}

	LOG_INF("Scan module initialized");
	return err;
}


static void auth_cancel(struct bt_conn *conn)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	LOG_INF("Pairing cancelled: %s", addr);
}


static void pairing_complete(struct bt_conn *conn, bool bonded)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	LOG_INF("Pairing completed: %s, bonded: %d", addr, bonded);
}


static void pairing_failed(struct bt_conn *conn, enum bt_security_err reason)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	LOG_WRN("Pairing failed conn: %s, reason %d", addr, reason);
}

static struct bt_conn_auth_cb conn_auth_callbacks = {
	.cancel = auth_cancel,
};

static struct bt_conn_auth_info_cb conn_auth_info_callbacks = {
	.pairing_complete = pairing_complete,
	.pairing_failed = pairing_failed
};

static void configure_gpio(void)
{
	int err;
	
#ifdef nRF21540DK
	nrf_gpio_cfg_output(13); 
	nrf_gpio_cfg_output(17); 
	nrf_gpio_cfg_output(15); 
	nrf_gpio_cfg_output(16);		
#elif BT840N_V4	
	#define APP_ANTSEL_PIN	20
	nrf_gpio_cfg_input(3,GPIO_PULL_UP);
	nrf_gpio_cfg_output(13); 
	nrf_gpio_cfg_output(17); 
	nrf_gpio_cfg_output(19); 
	nrf_gpio_cfg_output(32);	//P1.00 
	nrf_gpio_cfg_output(APP_ANTSEL_PIN); 		
#endif

#ifdef  nRF21540DK
	nrf_gpio_pin_clear(13);		//Led on
	nrf_gpio_pin_clear(14);		//Led on
	nrf_gpio_pin_clear(15);		//Led on
	nrf_gpio_pin_clear(16);		//Led on
        k_sleep(K_MSEC(300));
	nrf_gpio_pin_set(13);		//Led off
	nrf_gpio_pin_set(14);		//Led off
	nrf_gpio_pin_set(15);		//Led off
	nrf_gpio_pin_set(16);		//Led off
        k_sleep(K_MSEC(300));
	nrf_gpio_pin_clear(13);		//Led on
	nrf_gpio_pin_clear(14);		//Led on
	nrf_gpio_pin_clear(15);		//Led on
	nrf_gpio_pin_clear(16);		//Led on
        k_sleep(K_MSEC(300));
	nrf_gpio_pin_set(13);		//Led off
	nrf_gpio_pin_set(14);		//Led off
	nrf_gpio_pin_set(15);		//Led off
	nrf_gpio_pin_set(16);		//Led off
	k_sleep(K_MSEC(300));
	nrf_gpio_pin_clear(13);		//Led on
	nrf_gpio_pin_clear(14);		//Led on
	nrf_gpio_pin_clear(15);		//Led on
	nrf_gpio_pin_clear(16);		//Led on
        k_sleep(K_MSEC(300));
	nrf_gpio_pin_set(13);		//Led off
	nrf_gpio_pin_set(14);		//Led off
	nrf_gpio_pin_set(15);		//Led off
	nrf_gpio_pin_set(16);		//Led off	
#elif BT840N_V4
	nrf_gpio_pin_clear(13);		//Led on
	nrf_gpio_pin_clear(17);		//Led on
	nrf_gpio_pin_clear(19);		//Led on
	nrf_gpio_pin_clear(32);		//Led on,P1.00
        k_sleep(K_MSEC(300));
	nrf_gpio_pin_set(13);		//Led off
	nrf_gpio_pin_set(17);		//Led off
	nrf_gpio_pin_set(19);		//Led off
	nrf_gpio_pin_set(32);		//Led off,P1.00
        k_sleep(K_MSEC(300));
	nrf_gpio_pin_clear(13);		//Led on
	nrf_gpio_pin_clear(17);		//Led on
	nrf_gpio_pin_clear(19);		//Led on
	nrf_gpio_pin_clear(32);		//Led on,P1.00
        k_sleep(K_MSEC(300));
	nrf_gpio_pin_set(13);		//Led off
	nrf_gpio_pin_set(17);		//Led off
	nrf_gpio_pin_set(19);		//Led off
	nrf_gpio_pin_set(32);		//Led off,P1.00
        k_sleep(K_MSEC(300));
	//uint8_t rc = gpio_pin_get(led_dev,3);
	uint8_t rc = nrf_gpio_pin_read(3);
	if(rc == 0){		
        //gpio_pin_set(led_dev, APP_ANTSEL_PIN, 1);
		nrf_gpio_pin_clear(APP_ANTSEL_PIN);
		printk("Set ANTSEL_PIN(P0.20) low \n");
    }else{
        //gpio_pin_set(led_dev, APP_ANTSEL_PIN, 0);
		nrf_gpio_pin_set(APP_ANTSEL_PIN);
		printk("Set ANTSEL_PIN(P0.20) high \n");
    } 
#endif	
}

#ifdef client_sent_test
uint8_t hextoascii(uint8_t hex_byte)
{
    char result;
    if((hex_byte>=0)&&(hex_byte<=9))            
        result = hex_byte + 0x30;
    else if((hex_byte >= 10)&&(hex_byte <= 15)) 
        result = hex_byte + 0x37;
    else
        result = 0xff;
    return result;
}

void central_sample_event(struct k_timer *timer_id){

	if(is_connect)
	{
		Dummy_cnt++;
		if(Dummy_cnt > 0x39)
			Dummy_cnt = 0x30;
		Dummy_tx_buf[3]=Dummy_cnt;
		printk("Dummy_tx_buf = %c\n",Dummy_cnt);

	}else{
		return 0;
	}
		
	#if 0
		int err;
		//struct client_sent_buf_t *cs_temp;
		uint8_t	m_tx_buf[15];	
		uint16_t temp_len;	
		m_tx_buf[0]='+';
		m_tx_buf[1]='c';
		m_tx_buf[2]='l';
		m_tx_buf[3]='i';
		m_tx_buf[4]='e';
		m_tx_buf[5]='n';
		m_tx_buf[6]='t';
		m_tx_buf[7]='_';
		m_tx_buf[8]='R';
		m_tx_buf[9]='X';
		m_tx_buf[10]=':';
		m_tx_buf[11]=hextoascii((uint8_t)count_val%16);
		m_tx_buf[12]=0x0D;
		m_tx_buf[13]=0x0A;
		m_tx_buf[14]='\0';   			
		count_val++; 
		temp_len = temp_len+0x0F;
        if(count_val >255)
          count_val = 0; 
		if(is_connect){
			printk("m_tx_buf = %s\n",m_tx_buf);	
			//err = bt_nus_client_send(&nus_client, m_tx_buf, sizeof(temp_len));
		}
    #endif    
}

K_TIMER_DEFINE(my_timer, central_sample_event, NULL);
#endif

#ifdef MONITOR_TXPOWER_SETTING
void SWI3_IRQHandler(void)
{
    /* Clear event register */
    NRF_EGU3->EVENTS_TRIGGERED[0] = 0;
    LOG_INF("TX power: %d dBm", (int8_t)NRF_RADIO->TXPOWER);
}

void monitor_tx_power_register(void)
{
    IRQ_CONNECT(SWI3_IRQn, 5, SWI3_IRQHandler, NULL, 0);
    NVIC_EnableIRQ(SWI3_IRQn);
    NRF_EGU3->INTENSET = EGU_INTEN_TRIGGERED0_Msk;
    NRF_PPI->CH[0].EEP = (uint32_t) &NRF_RADIO->EVENTS_TXREADY;
    NRF_PPI->CH[0].TEP = (uint32_t) &NRF_EGU3->TASKS_TRIGGER[0];
    NRF_PPI->CHENSET = PPI_CHENSET_CH0_Msk;    
}
#endif

int main(void)
{
	int err;

	configure_gpio();

	err = bt_conn_auth_cb_register(&conn_auth_callbacks);
	if (err) {
		LOG_ERR("Failed to register authorization callbacks.");
		return 0;
	}

	err = bt_conn_auth_info_cb_register(&conn_auth_info_callbacks);
	if (err) {
		printk("Failed to register authorization info callbacks.\n");
		return 0;
	}

	err = bt_enable(NULL);
	if (err) {
		LOG_ERR("Bluetooth init failed (err %d)", err);
		return 0;
	}
	LOG_INF("Bluetooth initialized");

	if (IS_ENABLED(CONFIG_SETTINGS)) {
		settings_load();
	}

	err = uart_init();
	if (err != 0) {
		LOG_ERR("uart_init failed (err %d)", err);
		return 0;
	}

	err = scan_init();
	if (err != 0) {
		LOG_ERR("scan_init failed (err %d)", err);
		return 0;
	}

	err = nus_client_init();
	if (err != 0) {
		LOG_ERR("nus_client_init failed (err %d)", err);
		return 0;
	}

	printk("Starting Bluetooth Central UART example\n");

	#ifdef MONITOR_TXPOWER_SETTING
    monitor_tx_power_register();
	#endif
////
#ifdef nRF21540DK
	#define APP_MODE_PIN	17
	#ifdef APP_MODE_PIN
		nrf_gpio_cfg_output(APP_MODE_PIN);
		nrf_gpio_pin_set(APP_MODE_PIN);
	#endif
#elif BT840N_V4
	#define APP_MODE_PIN	39		//P1.07
	#ifdef APP_MODE_PIN
		//nrf_gpio_cfg_output(APP_MODE_PIN);
		//nrf_gpio_pin_clear(APP_MODE_PIN);
	#endif
#endif
////

	err = bt_scan_start(BT_SCAN_TYPE_SCAN_ACTIVE);
	if (err) {
		LOG_ERR("Scanning failed to start (err %d)", err);
		return 0;
	}

	LOG_INF("Scanning successfully started");
	printk("Version:v240_230703\n");
#ifdef client_sent_test
	#if 1
	//Timer setup
    //k_timer_init(&my_timer, central_sample_event, NULL);
    k_timer_start(&my_timer, K_MSEC(TIMER_INTERVAL_MSEC), K_MSEC(TIMER_INTERVAL_MSEC));		//start one shot timer that expires after 500 ms
	k_timer_status_sync(&my_timer);	
	#endif	
#endif	

			Dummy_tx_buf[0]='T';
			Dummy_tx_buf[1]='X';
			Dummy_tx_buf[2]=':';
			Dummy_tx_buf[3]=Dummy_cnt;
			Dummy_tx_buf[4]=0x0d;
			Dummy_tx_buf[5]=0x0a;

	for (;;) {

			err = bt_nus_client_send(&nus_client, Dummy_tx_buf, 6);
			if (err) {
				LOG_WRN("Failed to send data over BLE connection"
				"(err %d)", err);
			}
			err = k_sem_take(&nus_write_sem, NUS_WRITE_TIMEOUT);
			if (err) {
				LOG_WRN("NUS send timeout");
			}	

		
		
		#if 0
		/* Wait indefinitely for data to be sent over Bluetooth */
		struct uart_data_t *buf = k_fifo_get(&fifo_uart_rx_data,
						     K_FOREVER);

		err = bt_nus_client_send(&nus_client, buf->data, buf->len);
		if (err) {
			LOG_WRN("Failed to send data over BLE connection"
				"(err %d)", err);
		}

		err = k_sem_take(&nus_write_sem, NUS_WRITE_TIMEOUT);
		if (err) {
			LOG_WRN("NUS send timeout");
		}
		#endif		
	}
}
